from queue import deque
from copy import deepcopy

from sereia.utils import ConfigHandler
from sereia.keyword_match import KeywordMatch
from sereia.database import MongoHandler

from .candidate_network import CandidateNetwork


class CandidateNetworkHandler:
    def __init__(self, database_handler, **kwargs):
        self.database_handler = database_handler

    def generate_cns(self,schema_index,schema_graph,ranked_query_matches,keywords,weight_scheme,**kwargs):
        topk_cns = kwargs.get('topk_cns',20)

        returned_cns=[]
        num_cns_available = topk_cns

        for n, query_match in enumerate(ranked_query_matches):
            if topk_cns!=-1 and num_cns_available<=0:
                break

            cns_per_cur_qm = self.generate_cns_per_qm(schema_index,schema_graph,query_match,keywords,weight_scheme,**kwargs)

            for m, candidate_network in enumerate(cns_per_cur_qm):
                if num_cns_available<=0:
                    break

                returned_cns.append(candidate_network)

                num_cns_available -=1

        ranked_cns=sorted(
            returned_cns,
            key=lambda candidate_network: candidate_network.score,
            reverse=True
        )

        return ranked_cns

    def generate_cns_per_qm(self,schema_index,schema_graph,query_match,keywords,weight_scheme,**kwargs):
        max_cn_size = kwargs.get('max_cn_size',3)
        topk_cns_per_qm = kwargs.get('topk_cns_per_qm',1)
        
        instance_based_pruning = kwargs.get('instance_based_pruning',False)
        max_database_accesses  = kwargs.get('max_database_accesses',7)
        database_handler = self.database_handler

        directed_neighbor_sorting_function = kwargs.get('directed_neighbor_sorting_function',
                                                        self.factory_sum_norm_attributes(schema_index,weight_scheme))
        schema_prunning = kwargs.get('schema_prunning',True)
        desired_cn = kwargs.get('desired_cn',None)

        returned_cns = []
        empty_cns = []
        pruned_cns = set()

        def meet_pruning_conditions(jnkm):
            return not (
                #corretude rules
                jnkm.is_sound(schema_graph) and
                #checking whether it was already generated
                jnkm not in F and
                jnkm not in pruned_cns and
                jnkm not in returned_cns and
                jnkm not in empty_cns and
                # max cn size pruning
                len(jnkm)<=max_cn_size and
                (
                    not schema_prunning or
                    (
                        # max number of leaves pruning
                        jnkm.num_leaves() <= len(query_match) and
                        # max number of keyword-free matches pruning
                        jnkm.num_free_keyword_matches()+len(query_match) <= max_cn_size
                    )
                )
            )

        # JNKM stands for Joining Network of Keyword Matches
        cur_jnkm = CandidateNetwork()
        cur_jnkm.add_keyword_match(query_match.get_km_from_keyword(keywords[0]))

        if len(query_match)==1:
            cur_jnkm.calculate_score(query_match)
            returned_cns.append(cur_jnkm)
            return returned_cns

        table_hash={}
        for keyword_match in query_match:
            table_hash.setdefault(keyword_match.table,{KeywordMatch(keyword_match.table)}).add(keyword_match)

        F = deque()
        F.append(cur_jnkm)

        while F:
            cur_jnkm = F.popleft()

            for vertex_u in reversed(cur_jnkm.vertices()):
                keyword_match,_ = vertex_u

                sorted_directed_neighbors = sorted(
                    schema_graph.directed_neighbors(keyword_match.table),
                    key=directed_neighbor_sorting_function
                    )

                for direction,adj_table in sorted_directed_neighbors:

                    table_hash.setdefault(adj_table,{KeywordMatch(adj_table)})

                    for adj_keyword_match in table_hash[adj_table]:

                        if (adj_keyword_match not in cur_jnkm.keyword_matches() or
                            adj_keyword_match.is_free()):

                            next_jnkm = deepcopy(cur_jnkm)
                            next_jnkm.add_adjacent_keyword_match(vertex_u,adj_keyword_match,edge_direction=direction)

                            if not meet_pruning_conditions(next_jnkm):
                                # print(f'next_jnkm:\n{next_jnkm}')
                                if next_jnkm.is_total(query_match):
                                    if not next_jnkm.contains_keyword_free_match_leaf():                                      
                                        if not instance_based_pruning or not self.is_cn_empty(schema_graph,next_jnkm,handler=database_handler):
                                            next_jnkm.calculate_score(query_match)
                                            returned_cns.append(next_jnkm)
                                        else:
                                            empty_cns.append(next_jnkm)

                                        if desired_cn is not None and next_jnkm == desired_cn:
                                            return returned_cns
                                    else:
                                        #reduce JNKM
                                        # print(f'next_jnkm(pruned): non minimal\n{next_jnkm}')
                                        pruned_cns.add(next_jnkm)
                                        continue

                                    if (
                                        len(returned_cns)>=topk_cns_per_qm or
                                        (
                                            instance_based_pruning and
                                            len(returned_cns)+len(empty_cns)>=max_database_accesses
                                        )
                                    ):
                                        return returned_cns

                                elif len(next_jnkm)<max_cn_size:
                                    F.append(next_jnkm)
                                    # print(f'next_jnkm:\n{next_jnkm}')
                                else:
                                    pruned_cns.add(next_jnkm)
                            else:
                                pruned_cns.add(next_jnkm)                                
        return returned_cns

    def factory_sum_norm_attributes(self,schema_index,weight_scheme):

        def sum_norm_attributes(directed_neighbor):
            _,adj_table = directed_neighbor
            score1 = 0 
            if adj_table in schema_index:
                score1 = sum(
                    schema_index[adj_table][attribute]['norm'][weight_scheme]
                    for attribute in schema_index[adj_table]
                )
            score2 = adj_table
            return (-score1,score2)

        return sum_norm_attributes
    
    def is_cn_empty(self, schema_graph, candidate_network, handler='rel'):
        if handler == 'rel':
            query = candidate_network.get_sql_from_cn(schema_graph)

        if handler == 'mongo':
            query = candidate_network.get_mongo_query_from_cn(schema_graph)
            print(query)

        return True



# TODO: Functionalities needed for the index generation
# 1 - Traversal to obtain all the attributes from the document (in any level of nesting)
# 2 - Traversal to obtain all the values to index
from copy import deepcopy
import datetime

from bson.int64 import Int64

class DocumentTraverser:

    def __init__(self):
        self.traversed_attributes = {}
        self.document_attributes = {}
        self.indexable_content = []

    def _traverse(self, attribute_path, document_path, value, value_type):
        if value_type == list:
            self.document_attributes[document_path] = value_type
            for number, element in enumerate(value):
                updated_path = ''.join([
                    attribute_path,
                    '[',
                    str(number + 1),
                    ']',
                ])
                self._add_to_document_attributes(document_path, type(element))
                self._traverse(updated_path, document_path, element, type(element))
        elif value_type == dict:
            if document_path not in self.document_attributes:
                self.document_attributes[document_path] = value_type
            for attribute, attribute_value in value.items():
                updated_path = self._increment_document_path(attribute_path, attribute)
                updated_document_path = self._increment_document_path(document_path, attribute)
                self._add_to_document_attributes(updated_document_path, type(attribute_value))
                self._traverse(updated_path, updated_document_path, attribute_value, type(attribute_value))
        elif value_type in [str, int, float, Int64]:
            self._add_to_document_attributes(document_path, value_type)
            self.traversed_attributes[attribute_path] = value
            self.indexable_content.append(
                (document_path, value),
            )
        elif value_type == datetime.datetime:
            self._add_to_document_attributes(document_path, value_type)
            self.traversed_attributes[attribute_path] = value.strftime('%d/%m/%Y %H:%M:%S')

    def _add_to_document_attributes(self, document_path, value_type):
        if document_path not in self.document_attributes:
            self.document_attributes[document_path] = value_type

    def _increment_document_path(self, current_path, new_attribute):
        return '.'.join([current_path, new_attribute])
    
    def cleanup(self):
        self.traversed_attributes = {}
        # TODO: evaluate if we can mantain this dict while iterating through all documents
        # self.document_attributes = {}
        self.indexable_content = []

    def traverse(self, document):
        for attribute, value in document.items():
            # print(attribute, type(value))
            self._traverse(attribute, attribute, value, type(value))
    
    def get_traversed_attributes(self):
        return self.traversed_attributes

    def get_document_attributes(self):
        return self.document_attributes

    def get_document_attributes_without_top_level_attributes(self):
      filtered_document_attributes = deepcopy(self.document_attributes)
      attributes_for_removal = []
      for attribute in filtered_document_attributes:
          if '.' not in attribute and \
                  filtered_document_attributes[attribute] in [type(list()), type(dict())]:
              # print('Removing attribute from collection structure')
              attributes_for_removal.append(attribute)
          else:
              if filtered_document_attributes[attribute] in [type(list()), type(dict())]:
                  attributes_for_removal.append(attribute)
      for attribute in attributes_for_removal:
          del filtered_document_attributes[attribute]
      
      return filtered_document_attributes

    def get_indexable_content(self):
        return self.indexable_content


if __name__ == "__main__":
    document = {
  "_id": {
    "$oid": "52f587f0b42a75c4264022c4"
  },
  "teams": [
    {
      "name": "Washington Wizards",
      "abbreviation": "WAS",
      "score": 78,
      "home": True,
      "won": 0,
      "results": {
        "ast": 13,
        "blk": 2,
        "drb": 30,
        "fg": 27,
        "fg3": 4,
        "fg3_pct": ".211",
        "fg3a": 19,
        "fg_pct": ".375",
        "fga": 72,
        "ft": 20,
        "ft_pct": ".714",
        "fta": 28,
        "mp": 240,
        "orb": 9,
        "pf": 18,
        "plus_minus": "",
        "pts": 78,
        "stl": 8,
        "tov": 14,
        "trb": 39
      },
      "players": [
        {
          "ast": 1,
          "blk": 0,
          "drb": 6,
          "fg": 4,
          "fg3": 0,
          "fg3_pct": "",
          "fg3a": 0,
          "fg_pct": ".500",
          "fga": 8,
          "ft": 5,
          "ft_pct": ".833",
          "fta": 6,
          "mp": "35:00",
          "orb": 5,
          "pf": 4,
          "player": "Jared Jeffries",
          "plus_minus": -9,
          "pts": 13,
          "stl": 0,
          "tov": 2,
          "trb": 11
        },
        {
          "ast": 0,
          "blk": 0,
          "drb": 2,
          "fg": 5,
          "fg3": 1,
          "fg3_pct": ".250",
          "fg3a": 4,
          "fg_pct": ".500",
          "fga": 10,
          "ft": 4,
          "ft_pct": ".667",
          "fta": 6,
          "mp": "33:00",
          "orb": 1,
          "pf": 1,
          "player": "Jarvis Hayes",
          "plus_minus": "+2",
          "pts": 15,
          "stl": 1,
          "tov": 0,
          "trb": 3
        },
        {
          "ast": 3,
          "blk": 1,
          "drb": 2,
          "fg": 1,
          "fg3": 1,
          "fg3_pct": ".200",
          "fg3a": 5,
          "fg_pct": ".111",
          "fga": 9,
          "ft": 0,
          "ft_pct": "",
          "fta": 0,
          "mp": "30:00",
          "orb": 0,
          "pf": 0,
          "player": "Steve Blake",
          "plus_minus": "+4",
          "pts": 3,
          "stl": 1,
          "tov": 3,
          "trb": 2
        },
        {
          "ast": 2,
          "blk": 1,
          "drb": 5,
          "fg": 3,
          "fg3": 0,
          "fg3_pct": ".000",
          "fg3a": 1,
          "fg_pct": ".300",
          "fga": 10,
          "ft": 2,
          "ft_pct": "1.000",
          "fta": 2,
          "mp": "26:00",
          "orb": 0,
          "pf": 2,
          "player": "Larry Hughes",
          "plus_minus": "+7",
          "pts": 8,
          "stl": 0,
          "tov": 0,
          "trb": 5
        },
        {
          "ast": 2,
          "blk": 0,
          "drb": 4,
          "fg": 4,
          "fg3": 0,
          "fg3_pct": "",
          "fg3a": 0,
          "fg_pct": ".800",
          "fga": 5,
          "ft": 0,
          "ft_pct": "",
          "fta": 0,
          "mp": "18:00",
          "orb": 1,
          "pf": 2,
          "player": "Brendan Haywood",
          "plus_minus": "+2",
          "pts": 8,
          "stl": 0,
          "tov": 0,
          "trb": 5
        },
        {
          "ast": 0,
          "blk": 0,
          "drb": 5,
          "fg": 6,
          "fg3": 0,
          "fg3_pct": "",
          "fg3a": 0,
          "fg_pct": ".667",
          "fga": 9,
          "ft": 4,
          "ft_pct": ".667",
          "fta": 6,
          "mp": "30:00",
          "orb": 1,
          "pf": 3,
          "player": "Etan Thomas",
          "plus_minus": -18,
          "pts": 16,
          "stl": 3,
          "tov": 2,
          "trb": 6
        },
        {
          "ast": 1,
          "blk": 0,
          "drb": 1,
          "fg": 3,
          "fg3": 1,
          "fg3_pct": ".200",
          "fg3a": 5,
          "fg_pct": ".273",
          "fga": 11,
          "ft": 1,
          "ft_pct": ".500",
          "fta": 2,
          "mp": "21:00",
          "orb": 1,
          "pf": 2,
          "player": "Juan Dixon",
          "plus_minus": -18,
          "pts": 8,
          "stl": 1,
          "tov": 1,
          "trb": 2
        },
        {
          "ast": 2,
          "blk": 0,
          "drb": 2,
          "fg": 0,
          "fg3": 0,
          "fg3_pct": ".000",
          "fg3a": 3,
          "fg_pct": ".000",
          "fga": 5,
          "ft": 2,
          "ft_pct": ".500",
          "fta": 4,
          "mp": "20:00",
          "orb": 0,
          "pf": 2,
          "player": "Gilbert Arenas",
          "plus_minus": -25,
          "pts": 2,
          "stl": 1,
          "tov": 4,
          "trb": 2
        },
        {
          "ast": 0,
          "blk": 0,
          "drb": 2,
          "fg": 1,
          "fg3": 1,
          "fg3_pct": "1.000",
          "fg3a": 1,
          "fg_pct": ".250",
          "fga": 4,
          "ft": 0,
          "ft_pct": "",
          "fta": 0,
          "mp": "15:00",
          "orb": 0,
          "pf": 0,
          "player": "Mitchell Butler",
          "plus_minus": -18,
          "pts": 3,
          "stl": 1,
          "tov": 1,
          "trb": 2
        },
        {
          "ast": 2,
          "blk": 0,
          "drb": 1,
          "fg": 0,
          "fg3": 0,
          "fg3_pct": "",
          "fg3a": 0,
          "fg_pct": ".000",
          "fga": 1,
          "ft": 2,
          "ft_pct": "1.000",
          "fta": 2,
          "mp": "12:00",
          "orb": 0,
          "pf": 2,
          "player": "Lonny Baxter",
          "plus_minus": -7,
          "pts": 2,
          "stl": 0,
          "tov": 1,
          "trb": 1
        }
      ],
      "city": "Washington"
    }
  ],
  "date": {
    "$date": "2004-04-14T04:00:00.000+0000"
  }
}
    from pprint import pprint as pp
    from time import time
    start = time()
    d = DocumentTraverser() 
    d.traverse(document)
    # pp(d.get_indexable_content())
    pp(d.get_document_attributes())
    # pp(d.get_traversed_attributes())
    print('Total time: {}'.format(time() - start))



from copy import deepcopy
import shelve
import pickle


class SchemaIndex():
    def __init__(self):
        self._dict = {}

    def __iter__(self):
        yield from self._dict.keys()

    def __getitem__(self,word):
        return self._dict[word]

    def __setitem__(self,key,value):
        self._dict[key] = value

    def __repr__(self):
        return repr(self._dict)

    def __str__(self):
        return str(self._dict)

    def keys(self):
        yield from self.__iter__()

    def items(self):
        for key in self.__iter__():
            yield key, self.__getitem__(key)

    def values(self):
        for key in self:
            yield self[key]

    def create_entries(self,table_attributes,metrics):
        if type(table_attributes) == list:
            for table,attributes in table_attributes:
                self._dict.setdefault(table,{})
                for attribute in attributes:
                    self._dict[table].setdefault(attribute, deepcopy(metrics) )
        elif type(table_attributes) == dict:
            for table,attributes in table_attributes.items():
                self._dict.setdefault(table,{})
                for attribute in attributes:
                    self._dict[table].setdefault(attribute, deepcopy(metrics) )

    def tables_attributes(self, database_name, nested_structures=False):
        collection_structure_path = ''.join([
                'tmp/',
                database_name + '.pickle'
            ]
        )
        
        with open(collection_structure_path, 'rb') as f:
            collection_structure = pickle.load(
                f,
            )
        
        tables_attributes_set = set([])
        for table in self:
            for attribute in self[table]:
                if nested_structures:
                    tables_attributes_set.add((table, attribute))
                else:
                    if not collection_structure[table][attribute] in [type(dict()), type(list())]:
                        tables_attributes_set.add((table, attribute))

        return tables_attributes_set

    def get_num_total_attributes(self):
        return sum([len(attribute) for attribute in self.values()])

    def persist_to_file(self, file_directory, filename):
        filepath = file_directory + filename
        with shelve.open(filepath) as storage:
            for key,value in self._dict.items():
                storage[key]=value

    def load_from_file(self, filename):
        with shelve.open(filename,flag='r') as storage:
            for key,value in storage.items():
                self[key]=value



from collections import OrderedDict
import json
from os import listdir
from os.path import isfile, join

from nltk.featstruct import interactive_demo

from sereia.utils import ConfigHandler, next_path, last_path
from sereia.query_match import QueryMatch
from sereia.candidate_network import CandidateNetwork
from sereia.utils.document_traverser import DocumentTraverser
from sereia.utils.utils import DictToSTupleFormatter, load_datasets_ids_from_json


class EvaluationHandler:

    def __init__(self, dataset_name, config=None):
        self.config = config
        self.dataset_name = dataset_name

    def load_golden_standards(self):
        golden_standards_path = self.config.queryset_filepath
        with open(golden_standards_path, mode='r') as f:
            data = json.load(f)

        golden_standards = OrderedDict()
        for item in data:
            if 'query_matches' in item:
                item['query_matches'] = [QueryMatch.from_json_serializable(json_serializable_qm)
                                         for json_serializable_qm in item['query_matches']]

            if 'candidate_networks' in item:
                item['candidate_networks'] = [CandidateNetwork.from_json_serializable(json_serializable_cn)
                                         for json_serializable_cn in item['candidate_networks']]

            golden_standards[item['keyword_query']] = item

        self.golden_standards = golden_standards

    def load_results_from_file(self,**kwargs):
        approach = kwargs.get('approach','standard')
        results_filename = kwargs.get(
            'results_filename',
            last_path(f'{self.config.results_directory}{self.config.queryset_name}-{approach}-%03d.json')
        )

        with open(results_filename,mode='r') as f:
            data = json.load(f)
        return data


    def evaluate_results(self,results, **kwargs):
        '''
        results_filename is declared here for sake of readability. But the
        default value is only set later to get a timestamp more accurate.
        '''
        results_filename = kwargs.get('results_filename',None)
        approach = kwargs.get('approach','standard')
        skip_ranking_evaluation = kwargs.get('skip_ranking_evaluation',False)
        skip_document_retrieval_evaluation = kwargs.get('skip_document_retrieval_evaluation', False)
        write_evaluation_only  = kwargs.get('write_evaluation_only',False)

        if not skip_ranking_evaluation:
            self.evaluate_query_matches(results)
            self.evaluate_candidate_networks(results)

        if not skip_document_retrieval_evaluation:
            self.evaluate_retrieved_documents(results)

        self.evaluate_performance(results)
        self.evaluate_num_schema_keyword_matches(results)
        self.evaluate_num_value_keyword_matches(results)
        self.evaluate_num_keyword_matches(results)
        self.evaluate_num_query_matches(results)
        self.evaluate_num_candidate_networks(results)
        self.evaluate_keyword_matches(results)

        if write_evaluation_only:
            del results['results']

        if results_filename is None:
            results_filename = next_path(f'{self.config.results_directory}{self.config.queryset_name}-{approach}-%03d.json')
        
        with open(results_filename,mode='w') as f:
            json.dump(results,f, indent = 4)

        return results

    def evaluate_query_matches(self,results,**kwargs):
        max_k = kwargs.get('max_k',10)

        results.setdefault('evaluation',{})
        results['evaluation']['query_matches']={}

        relevant_positions = []
        for item in results['results']:

            if 'query_matches' in item and 'query_matches' in self.golden_standards[item['keyword_query']]:
                if len(self.golden_standards[item['keyword_query']]['query_matches']) <= 0:
                    continue

                golden_qm = self.golden_standards[item['keyword_query']]['query_matches'][0]

                query_matches = [QueryMatch.from_json_serializable(json_serializable_qm)
                                 for json_serializable_qm in item['query_matches']]

                relevant_position = self.get_relevant_position(query_matches,golden_qm)

                relevant_positions.append(relevant_position)

        results['evaluation']['query_matches']['mrr'] = self.get_mean_reciprocal_rank(relevant_positions)

        precision_at_k = {f'p@{k}' : self.get_precision_at_k(k,relevant_positions)
                          for k in range(1,max_k+1)}
        results['evaluation']['query_matches'].update(precision_at_k)

        results['evaluation']['query_matches']['relevant_positions']=relevant_positions

        print('QM Evaluation {}'.format(results['evaluation']['query_matches']))

    def evaluate_candidate_networks(self,results,**kwargs):
        max_k = kwargs.get('max_k',10)

        results.setdefault('evaluation',{})
        results['evaluation']['candidate_networks']={}

        relevant_positions = []
        for item in results['results']:
            if 'candidate_networks' in item and len(item['candidate_networks'])  and 'candidate_networks' in self.golden_standards[item['keyword_query']]:
                if len(self.golden_standards[item['keyword_query']]['candidate_networks']) <= 0:
                    continue
                
                golden_cn = self.golden_standards[item['keyword_query']]['candidate_networks'][0]

                candidate_networks = [CandidateNetwork.from_json_serializable(json_serializable_cn)
                                 for json_serializable_cn in item['candidate_networks']]

                relevant_position = self.get_relevant_position(candidate_networks,golden_cn)

                relevant_positions.append(relevant_position)

        results['evaluation']['candidate_networks']['mrr'] = self.get_mean_reciprocal_rank(relevant_positions)

        precision_at_k = {f'p@{k}' : self.get_precision_at_k(k,relevant_positions)
                          for k in range(1,max_k+1)}
        results['evaluation']['candidate_networks'].update(precision_at_k)

        results['evaluation']['candidate_networks']['relevant_positions']=relevant_positions

        print('CN Evaluation {}'.format(results['evaluation']['candidate_networks']))

    def evaluate_retrieved_documents(self, results):

        results['evaluation']['retrieval_score'] = []
        GROUND_TRUTH_FOLDER = './ground_truth/{}/'.format(self.dataset_name)

        traverser = DocumentTraverser()
        formatter = DictToSTupleFormatter()

        mongo_id_as_id = False
        databases_collection_ids = load_datasets_ids_from_json()
        database_collections = databases_collection_ids[self.dataset_name]

        for collection in database_collections:
            if '_id' in database_collections[collection]:
                mongo_id_as_id = True

        for item in results['results']:
            retrieved_documents = item['retrieved_documents']
            formatted_retrieved_documents = []
            retrieved_documents_tuples = set([])
            for document in retrieved_documents:
                if not mongo_id_as_id:
                    del document['mongo_id']
                traverser.traverse(document)
                traversed = traverser.get_traversed_attributes()
                formatted_retrieved_documents.append(
                    traversed,
                )
                retrieved_documents_tuples.add(
                    formatter.format(traversed)
                )
                traverser.cleanup()
            
            ground_truth_file = GROUND_TRUTH_FOLDER + 'gt_query_{}.json'.format(item['query_number'] + 1)
            with open(ground_truth_file) as f:
                ground_truth = json.load(f)
            
            query_ground_truth = set([])
            for document in ground_truth['expected_documents']:
                if not mongo_id_as_id:
                    del document['mongo_id']
                query_ground_truth.add(
                    formatter.format(document)
                )

            intersection_len = len(retrieved_documents_tuples.intersection(query_ground_truth))
            union_len = len(retrieved_documents_tuples.union(query_ground_truth))
            retrieved_len = len(retrieved_documents)

            if not retrieved_len:
                precision = 0
            else:
                precision = intersection_len / retrieved_len
            
            if len(query_ground_truth):
                recall = intersection_len / len(query_ground_truth)
            else:
                recall = 0

            results['evaluation']['retrieval_score'].append(
                {
                    'keyword_query': item['keyword_query'],
                    'precision': precision,
                    'recall': recall,
                    'num_documents_retrieved': retrieved_len,
                    'num_documents_expected': len(query_ground_truth)
                }
            )

            del item['retrieved_documents']


    def evaluate_performance(self,results,**kwargs):
        results.setdefault('evaluation',{})
        results['evaluation']['performance']={}

        for item in results['results']:

            if 'elapsed_time' in item:
                for phase in item['elapsed_time']:
                    results['evaluation']['performance'].setdefault(phase,[]).append(item['elapsed_time'][phase])

    def evaluate_num_schema_keyword_matches(self, results, **kwargs):
        results.setdefault('evaluation', {})
        results['evaluation']['num_schema_keyword_matches'] = []

        for item in results['results']:

            if 'num_schema_keyword_matches' in item:
                results['evaluation']['num_schema_keyword_matches'].append(
                    item['num_schema_keyword_matches'])

    def evaluate_num_value_keyword_matches(self, results, **kwargs):
        results.setdefault('evaluation', {})
        results['evaluation']['num_value_keyword_matches'] = []

        for item in results['results']:

            if 'num_value_keyword_matches' in item:
                results['evaluation']['num_value_keyword_matches'].append(
                    item['num_value_keyword_matches'])

    def evaluate_num_keyword_matches(self,results,**kwargs):
        results.setdefault('evaluation',{})
        results['evaluation']['num_keyword_matches']=[]

        for item in results['results']:

            if 'num_keyword_matches' in item:
                results['evaluation']['num_keyword_matches'].append(item['num_keyword_matches'])

    def evaluate_num_query_matches(self,results,**kwargs):
        results.setdefault('evaluation',{})
        results['evaluation']['num_query_matches']=[]

        for item in results['results']:

            if 'num_query_matches' in item:
                results['evaluation']['num_query_matches'].append(item['num_query_matches'])
    
    def evaluate_keyword_matches(self,results,**kwargs):
        results.setdefault('evaluation',{})
        results['evaluation']['keyword_matches']=[]

        for item in results['results']:

            if 'keyword_matches' in item:
                results['evaluation']['keyword_matches'].append(str(item['keyword_matches']))

    def evaluate_num_candidate_networks(self,results,**kwargs):
        results.setdefault('evaluation',{})
        results['evaluation']['num_candidate_networks']=[]
        for item in results['results']:
            if 'num_candidate_networks' in item:
                results['evaluation']['num_candidate_networks'].append(item['num_candidate_networks'])

    def get_relevant_position(self,items,ground_truth):
        for i,item in enumerate(items):
            if item==ground_truth:
                return (i+1)
        return -1

    def get_mean_reciprocal_rank(self,relevant_positions):
        if len(relevant_positions)==0:
            return 0
        sum = 0
        for relevant_position in relevant_positions:
            if relevant_position != -1:
                reciprocal_rank = 1/relevant_position
            else:
                reciprocal_rank = 0
            sum += reciprocal_rank

        mrr = sum/len(relevant_positions)
        return mrr

    def get_precision_at_k(self,k,relevant_positions):
        if len(relevant_positions)==0:
            return 0
        sum = 0
        for relevant_position in relevant_positions:
            if relevant_position <= k and relevant_position!=-1:
                sum+=1
        presition_at_k = sum/len(relevant_positions)
        return presition_at_k



import itertools
import re
from timeit import default_timer as timer
from pprint import pprint as pp

from nltk.stem import WordNetLemmatizer

from sereia.utils import ConfigHandler, Similarity
from .keyword_match import KeywordMatch


class KeywordMatchHandler:
    def __init__(self, similarity):
        self.similarities = similarity

    def get_keyword_matches(self, keywords, value_index, schema_index,**kwargs):
        sk_matches = self.schema_keyword_match_generator(keywords, schema_index)
        vk_matches = self.value_keyword_match_generator(keywords, schema_index)
        return sk_matches+vk_matches

    def value_keyword_match_generator(self, Q,value_index,**kwargs):

        ignored_collections = kwargs.get('ignored_collections',[])
        ignored_attributes = kwargs.get('ignored_attributes',[])

        #Input:  A keyword query Q=[k1, k2, . . . , km]
        #Output: Set of non-free and non-empty tuple-sets Rq

        '''
        The tuple-set Rki contains the tuples of Ri that contain all
        terms of K and no other keywords from Q
        '''

        #Part 1: Find sets of tuples containing each keyword
        P = {}
        start = timer()
        for keyword in Q.get_parsed_value():

            if isinstance(keyword, list):
                for term in keyword:
                    if not self.is_keyword_stored_in_index(term, value_index):
                        continue
                
                self.generate_compound_value_keyword_match(
                    keyword,
                    value_index,
                    P,
                    ignored_collections,
                    ignored_attributes,
                )

                continue
            elif not self.is_keyword_stored_in_index(keyword, value_index):
                    continue

            for table in value_index[keyword]:
                if table in ignored_collections:
                    continue

                for (attribute,ctids) in value_index[keyword][table].items():
                    if (table,attribute) in ignored_attributes:
                        continue

                    vkm = KeywordMatch(table, value_filter={attribute:{keyword}})
                    P[vkm] = set(ctids)

        #Part 2: Find sets of tuples containing larger termsets
        P = self.disjoint_itemsets(P)

        # print('Total time: {}'.format((timer() - start)* 1000))

        #Part 3: Ignore tuples
        return list(P)

    def generate_compound_value_keyword_match(self, compound_query_segment, value_index, document_set, ignored_collections, ignored_attributes):
        matches_data = {}
        keyword_to_attributes = {}
        attributes_intersection = set([])

        for keyword in compound_query_segment:
            if keyword not in keyword_to_attributes:
                keyword_to_attributes[keyword] = set([])
            for collection in value_index[keyword]:
                if collection in ignored_collections:
                    continue

                if collection not in matches_data:
                    matches_data[collection] = {}

                for (attribute, ctids) in value_index[keyword][collection].items():
                    if (collection, attribute) in ignored_attributes:
                        continue

                    keyword_to_attributes[keyword].add(attribute)

                    if attribute not in matches_data[collection]:
                        # TODO:
                        matches_data[collection][attribute] = set(ctids)
                        continue

                    matches_data[collection][attribute] = matches_data[collection][attribute].intersection(
                        set(ctids),
                    )

        for keyword in keyword_to_attributes:
            if len(attributes_intersection) == 0:
                attributes_intersection = keyword_to_attributes[keyword]
                continue

            attributes_intersection = attributes_intersection.intersection(
                keyword_to_attributes[keyword])

        for collection in matches_data:
            for attribute in matches_data[collection]:
                if attribute not in attributes_intersection or len(matches_data[collection][attribute]) == 0:
                    print('attribute not in intersection or empty: {}'.format(attribute))
                    continue
                vkm = KeywordMatch(collection, value_filter={
                                    attribute: {frozenset(compound_query_segment)}})
                document_set[vkm] = matches_data[collection][attribute]

    def disjoint_itemsets(self, pool):
        #Input: A Set of non-empty tuple-sets for each keyword alone P
        #Output: The Set P, but now including larger termsets (process Intersections)

        '''
        Termset is any non-empty subset K of the terms of a query Q
        '''
        
        delayed_removal = {}
        next_stage_pool = {}
        
        for ( vkm_i , vkm_j ) in itertools.combinations(pool,2):
            if vkm_i.table == vkm_j.table:
                joint_tuples = pool[vkm_i] & pool[vkm_j]

                if len(joint_tuples)>0:

                    joint_predicates = {}

                    for attribute, keywords in vkm_i.value_filter:
                        joint_predicates.setdefault(attribute,set()).update(keywords)

                    for attribute, keywords in vkm_j.value_filter:
                        joint_predicates.setdefault(attribute,set()).update(keywords)

                    vkm_ij = KeywordMatch(vkm_i.table,value_filter=joint_predicates)
                    next_stage_pool[vkm_ij] = joint_tuples

                    delayed_removal.setdefault(vkm_i,set()).update(joint_tuples)
                    delayed_removal.setdefault(vkm_j,set()).update(joint_tuples)

        for vkm_k in delayed_removal:
            tuples_to_remove = delayed_removal[vkm_k]
            pool[vkm_k].difference_update(tuples_to_remove)
            if len(pool[vkm_k])==0:
                del pool[vkm_k]

        if len(next_stage_pool)>0:
            pool.update(self.disjoint_itemsets(next_stage_pool))
        return pool

    def schema_keyword_match_generator(self, Q, schema_index,**kwargs):
        ignored_collections = kwargs.get('ignored_collections',[])
        ignored_attributes = kwargs.get('ignored_attributes',[])
        threshold = kwargs.get('threshold',1)
        keyword_matches_to_ignore = kwargs.get('keyword_matches_to_ignore',set())

        S = []
        start_skm_time = timer()
        lemmatizer = WordNetLemmatizer()
        for keyword in Q.get_parsed_value():
            if isinstance(keyword, list):
                continue

            for table in schema_index:
                if table in ignored_collections:
                    continue

                for attribute in itertools.chain('*', schema_index[table].keys() ):

                    if (
                        attribute=='id' or
                        (table,attribute) in ignored_attributes or
                        attribute==table
                        ):
                        continue

                    leaf_attribute = attribute.split('.')[-1]
                    lemmatized_attribute = lemmatizer.lemmatize(keyword)
                    if attribute == '*' and lemmatized_attribute == table:
                        sim = 1.0
                    elif lemmatized_attribute == leaf_attribute:
                        sim = 1.0
                    else:
                        sim = 0.0

                    if sim >= threshold:
                        skm = KeywordMatch(table,schema_filter={attribute:{keyword}})
                        if skm not in keyword_matches_to_ignore:
                            S.append(skm)
            
            # print('Total SKM time: {}'.format(timer() - start_skm_time))
        return S

    def filter_kwmatches_by_compound_keywords(self, vk_matches, compound_keywords):
        '''
        Value-keyword matches which contains only part of a compound_keyword are
        pruned.
        '''
        if len(compound_keywords)==0:
            return vk_matches

        filtered_vk_matches = set()
        for value_keyword_match in vk_matches:
            must_remove = False
            set_a = set(value_keyword_match.keywords())

            for compound_keyword in compound_keywords:
                set_b = set(compound_keyword)

                set_ab = set_a | set_b
                if len(set_ab)>0 or len(set_ab)<len(set_a):
                    must_remove = True

            if not must_remove:
                filtered_vk_matches.add(value_keyword_match)

        return filtered_vk_matches

    def is_keyword_stored_in_index(self, keyword, index):
        if keyword in index:
            return True
        
        return False



from pathlib import Path
import pickle
from pprint import pprint as pp

import pymongo

from sereia.database.mongo_iter import MongoIter
from sereia.utils import ConfigHandler, DocumentTraverser


class MongoHandler:
    def __init__(self, database_name, database_credentials, config_handler, **kwargs):
        self.database_name = database_name
        self.database_credentials = database_credentials
        self.database_client = pymongo.MongoClient(
            database_credentials,
        )

        self.database_batch_cursor_size = kwargs.get('database_batch_cursor_size', 10000)

    def set_database_name(self, database_name):
        self.database_name = database_name

    def execute(self, base_collection, query):
        database = self.database_client[
            self.database_name
        ]

        mongo_result = database[base_collection].aggregate(
            query,
            allowDiskUse=True,
        )

        return mongo_result

    def get_collections_and_attributes(self, indexable_attributes=None):
        database = self.database_client[self.database_name]

        collections_attributes = {}
        collections_structure = {}
        for collection in self.get_collections_names():
            collections_attributes[collection] = set([])
            collections_structure[collection] = {}

            projection_attributes = None

            if indexable_attributes:
                for item in indexable_attributes:
                    if item[0] == collection:
                        projection_attributes = item[1]

            for document in database[collection].find(
                projection=projection_attributes,
                batch_size=self.database_batch_cursor_size,
            ):

                traverser = DocumentTraverser()
                traverser.traverse(document)
                document_attributes = traverser.get_document_attributes().keys()
                for attribute in document_attributes:

                    if attribute != '_id':
                        collections_attributes[collection].add(
                            attribute,
                        )

                attributes_with_types = traverser.get_document_attributes()

                for attribute in attributes_with_types:
                    if attributes_with_types[attribute] != type(None):
                        collections_structure[collection][attribute] = attributes_with_types[attribute]

            collections_attributes[collection] = list(
                collections_attributes[collection]
            )

        print(f'Storing {self.database_name} database structure...')
        Path('tmp/').mkdir(exist_ok=True)
        with open(
            '/'.join(['tmp/', self.database_name + '.pickle']),
            mode = 'wb',
        ) as f:
            pickle.dump(
                collections_structure,
                f,
                protocol=pickle.HIGHEST_PROTOCOL
            )

        return collections_attributes

    def iterate_over_keywords(self, schema_index, indexable_attributes, **kwargs):
        schema_index_attributes = schema_index.tables_attributes(self.database_name, nested_structures=False)
        if indexable_attributes:
            for attribute in indexable_attributes:
                if attribute not in schema_index_attributes:
                    print('Attribute {} not located in Schema Index... Re-check and try again.'.format(attribute))
                    return None

        return MongoIter(
            indexable_attributes,
            self.database_name,
            self.database_client,
            self.database_batch_cursor_size,
            **kwargs
        )

    def exist_results(self, query):
        database = self.database_client[self.database_name]
        collection, mongo_query = query.build()

        count = database[collection].aggregate(mongo_query)

        for item in count:
            if item:
                count.close()
                return True

        return False
    
    def get_databases(self):
        databases = self.database_client.list_database_names()
        databases = [database for database in databases if database not in ['admin', 'local']]

        return databases

    def get_collections_names(self):
        database = self.database_client[self.database_name]
        filter = {"name": {"$regex": r"^(?!system\.)"}}

        collection_names = database.list_collection_names(filter=filter)

        return collection_names



import  itertools
from sereia.keyword_match import KeywordMatch

from .query_match import QueryMatch


class QueryMatchHandler:

    def generate_query_matches(self, keywords, keyword_matches, **kwargs):
        #Input:  A keyword query Q, The set of non-empty non-free tuple-sets Rq
        #Output: The set Mq of query matches for Q
        max_qm_size = kwargs.get('max_qm_size',3)
        query_matches = []

        for i in range(1,min(len(keywords), max_qm_size)+1):
            for candidate_query_match in itertools.combinations(keyword_matches,i):
                if self.has_minimal_cover(candidate_query_match,keywords):
                    merged_query_match = self.merge_schema_filters(candidate_query_match)
                    query_match = QueryMatch(merged_query_match)
                    query_matches.append(query_match)

        return query_matches

    def has_minimal_cover(self, candidate_query_match, keywords):
        #Input:  A subset CM (Candidate Query Match) to be checked as total and minimal cover
        #Output: If the match candidate is a TOTAL and MINIMAL cover

        # Check whether it is total

        if len({keyword
            for keyword_match in candidate_query_match
            for keyword in keyword_match.keywords()
            }) != len(set(keywords)):

            return False

        # Check whether it is minimal
        for element in candidate_query_match:
            if len({keyword
                for keyword_match in candidate_query_match
                for keyword in keyword_match.keywords()
                if keyword_match!=element
                }) == len(set(keywords)):

                return False

        return True

    def merge_schema_filters(self, query_matches):
        table_hash={}
        for keyword_match in query_matches:
            joint_schema_filter,value_keyword_matches = table_hash.setdefault(keyword_match.table,({},set()))

            for attribute, keywords in keyword_match.schema_filter:
                joint_schema_filter.setdefault(attribute,set()).update(keywords)

            if len(keyword_match.value_filter) > 0:
                value_keyword_matches.add(keyword_match)

        merged_qm = set()
        for table,(joint_schema_filter,value_keyword_matches) in table_hash.items():
            if len(value_keyword_matches) > 0:
                joint_value_filter = {attribute:keywords
                                    for attribute,keywords in value_keyword_matches.pop().value_filter}
            else:
                joint_value_filter={}

            joint_keyword_match = KeywordMatch(table,
                                               value_filter=joint_value_filter,
                                               schema_filter=joint_schema_filter)

            merged_qm.add(joint_keyword_match)
            merged_qm.update(value_keyword_matches)

        return merged_qm

    def rank_query_matches(self, query_matches, value_index, schema_index, similarity, weight_scheme):
        for query_match in query_matches:
            query_match.calculate_total_score(
                value_index,
                schema_index,
                similarity,
                weight_scheme,
            )

        ranked_query_matches = sorted(query_matches,key=lambda query_match: query_match.total_score,reverse=True)
    
        return ranked_query_matches



from collections import namedtuple
from json import load as json_load
from pathlib import Path
import pickle

class DictToSTupleFormatter(object):

    _NAME = 'STuple'

    def format(self, dictionary):
        sorted_keys = sorted(dictionary.keys())
        sorted_values = []
        
        for key in sorted_keys:
            sorted_values.append(dictionary[key])

        formatted_keys = [
            k.split('.')[1] if '.' in k else k for k in sorted_keys
        ]

        try:
            stuple = namedtuple(
                self._NAME,
                formatted_keys,
            )
        except ValueError as e:
            keys_set = set([])
            for key in formatted_keys:
                keys_set.add(key)
            stuple = namedtuple(
                self._NAME,
                keys_set,
            )
            sorted_values = []
            for key in keys_set:
                sorted_values.append(dictionary[key])

        generated_stuple = stuple(
            *sorted_values,
        )

        return generated_stuple

class AttributeTreeGenerator(dict):
    __slots__ = ['_dict', '_single_length_attribute_tree']

    def __init__(self):
        self._dict = {}
        self._single_length_attribute_tree = []
    
    def add_attribute(self, attribute, value):
        attribute_path = attribute.split('.')
        if len(attribute_path) == 1:
            self._single_length_attribute_tree.append(('', [(attribute, value,)]))
        else:
            leaf_attribute = attribute_path[-1]
            path = '.'.join(
                attribute_path[:-1],
            )

            if path not in self._dict:
                self._dict[path] = [
                    (leaf_attribute, value),
                ]
            else:
                self._dict[path].append((leaf_attribute, value,))

    def generate(self):
        attribute_trees = []
        for root in self._dict:
            tree = (
                root,
                self._dict[root],
            )
            attribute_trees.append(tree)
        
        for tree in self._single_length_attribute_tree:
            attribute_trees.append(tree)

        return attribute_trees

    def __repr__(self):
        return self._dict

def modify_retrieved_documents(retrieved_documents):
    modified_documents = []
    while len(retrieved_documents) > 0:
        document = retrieved_documents.pop()
        if '_id' in document:
            document['mongo_id'] = str(document['_id'])
        del document['_id']
        modified_documents.append(document)

    return modified_documents

def load_schema_graph_from_json(database):
    PATH = f'./schema_graphs/{database}/{database}_schema_graph.json'
    schema_graph_file_path_exists = Path(PATH).is_file()
    schema_graph = {}

    if schema_graph_file_path_exists:
        with open(PATH) as f:
            schema_graph = json_load(f)
        return schema_graph

    return Exception("Schema Graph file not found")

def load_datasets_ids_from_json():
    PATH = f'./collections_ids/datasets_collection_ids.json'
    datasets_ids_file_exists = Path(PATH).is_file()

    if datasets_ids_file_exists:
        with open(PATH) as f:
            return json_load(f)
    
    raise Exception("Datasets collection IDs file not found")

def load_database_structure_from_file(database_name):
    PATH = f'tmp/{database_name}.pickle'
    with open(PATH, 'rb') as f:
        database_structure = pickle.load(
            f,
        )
    
    return database_structure

if __name__ == '__main__':
    from pprint import pprint as pp

    tree = AttributeTreeGenerator()
    tree.add_attribute('teams.players.player', 'jon')
    tree.add_attribute('teams.players.player', 'jon')
    tree.add_attribute('teams.players.ast', 10)
    pp(tree.generate())




from pathlib import Path

from sereia.index.index_handler import IndexHandler
from sereia.database import MongoHandler
from sereia.utils import ConfigHandler


if __name__ == "__main__":
    database = 'imdb'
    print(f'Indexing database {database}')
    
    # TODO: check if the tmp/datasets/ folder path exist
    index_folder = './tmp/datasets/{}/'.format(
        database,
    )
    Path(index_folder).mkdir(parents=True, exist_ok=True)

    dataset_config_filepath = f'./datasets_config/{database}.json'
    config = ConfigHandler(
        config_directory='./config/',
        reset=True, dataset_config_filepath=dataset_config_filepath)
    config.set_logging_level('INFO')
    
    indexHandler = IndexHandler(
        MongoHandler(
            database,
        ), 
        config=config,
    )
    indexHandler.create_indexes()



import re
from string import punctuation,ascii_uppercase,ascii_lowercase
from nltk.corpus import stopwords
from nltk.tokenize import word_tokenize
from nltk.stem import WordNetLemmatizer


class Tokenizer:
    stop_words = set(stopwords.words('english'))-{'will'}
    re_spaces = re.compile(r'\s+')
    translate_table = str.maketrans(ascii_uppercase+punctuation,ascii_lowercase+' '*len(punctuation))
    wlm = WordNetLemmatizer()
    del translate_table[ord('_')]

    def __init__(self, **kwargs):
        self.tokenize_method = kwargs.get('tokenize_method','simple')

    def compound_keywords(self,keyword_query):
        return [self.tokenize(segment) for segment in keyword_query.split('"')[1::2]]

    def keywords(self,keyword_query):
        return self.tokenize(keyword_query)
    
    def text_translator(self, text):
        keywords = []
        split_text = text.split(' ')
        for word in split_text:
            # translate_table = word.maketrans(punctuation, len(punctuation) * ' ')
            # word = word.translate(translate_table)
            # word = word.strip()
            word = word.replace(',', '')
            # word = word.strip('.')
            word = word.strip()
            keywords.append(word.lower())
        
        return keywords

    def tokenize(self,text):
        if self.tokenize_method == 'simple':
            return [keyword
                    for keyword in self.text_translator(str(text))
                    if keyword not in Tokenizer.stop_words
            ]

        if self.tokenize_method == 'nltk':
            word_tokens = word_tokenize(text)
            keywords = [w for w in word_tokens if not w in self.stop_words]
            return keywords

    def extract_keywords(self, keyword_query):
        compound_operators_indexes = []
        valid_keyword_query = []
        compound_keyword_positions = []

        keywords = keyword_query.split(' ')
        keyword_query_length = len(keywords)

        for keyword_position, keyword in enumerate(keywords):
            keyword = keyword.lower()
            keyword = keyword.replace(',', '')
            keyword = keyword.strip()

            if keyword in Tokenizer.stop_words:
                continue

            if keyword == '+':
                if (keyword_position - 1) in compound_keyword_positions:
                    raise Exception(
                        'Unable to process query due to multiple operands for compound operator')
                
                if (keyword_position - 1) < 0 or\
                        (keyword_position + 1) > (keyword_query_length - 1):
                    raise Exception('Missing operand for compound operator')

                compound_operators_indexes.append(
                    keyword_position,
                )

                valid_keyword_query.append(
                    [keywords[keyword_position - 1],
                        keywords[keyword_position + 1]],
                )

                compound_keyword_positions.extend(
                    [keyword_position - 1, keyword_position + 1],
                )

                continue

            if keyword_position not in compound_keyword_positions:
                if (keyword_position + 1) < keyword_query_length and keywords[keyword_position + 1] == '+':
                    continue

                valid_keyword_query.append(
                    keyword,
                )

        return valid_keyword_query

if __name__ == "__main__":
    t = Tokenizer()

    q = t.extract_keywords(
        "ariel + antony afonso"
    )

    print(q)





import json
import string
import re
from pprint import pprint as pp

class KeywordMatch:

    def __init__(self, table, value_filter={},schema_filter={}):
        self.__slots__ =['table','schema_filter','value_filter']
        self.table = table
        self.schema_filter= frozenset({ (attribute,frozenset(keywords)) for attribute,keywords in schema_filter.items()})
        self.value_filter= frozenset({ (attribute,frozenset(keywords)) for attribute,keywords in value_filter.items()})

    def is_free(self):
        return len(self.schema_filter)==0 and len(self.value_filter)==0

    def schema_mappings(self):
        for attribute, keywords in self.schema_filter:
            yield (self.table,attribute,keywords)

    def value_mappings(self):
        for attribute, keywords in self.value_filter:
            yield (self.table,attribute,keywords)
    
    def has_value_mappings(self):
        return self.value_filter != frozenset({})

    def mappings(self):
        for attribute, keywords in self.schema_filter:
            yield ('s',self.table,attribute,keywords)
        for attribute, keywords in self.value_filter:
            yield ('v',self.table,attribute,keywords)

    def keywords(self,schema_only=False):
        for _, keywords in self.schema_filter:
            yield from keywords
        if not schema_only:
            for _, keywords in self.value_filter:
                for keyword in keywords:
                    if isinstance(keyword, frozenset):
                        for term in keyword:
                            yield term
                    else:
                        yield keyword

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        def str_filter(filter_type,fltr):
            if len(fltr) == 0:
                return ""
            
            str_repr = []
            
            for attribute, keywords in fltr:
                for keyword in keywords:
                    if isinstance(keyword, frozenset):
                        compound_keyword_segments = [
                            '{}'.format(
                                ','.join(keyword)
                            )
                        ]

                        compound_keyword_segments = ','.join(
                            compound_keyword_segments
                        )

                        str_repr.append(
                            (attribute, compound_keyword_segments),
                        )
                    else:
                        str_repr.append(
                            (
                                attribute,
                                keyword,
                            )
                        )

            return ".{}({})".format(
                filter_type,
                ','.join(
                    [
                        "{}{{{}}}".format(
                            attribute,
                            representation
                        ) for attribute, representation in str_repr
                    ]
                )
            )

        return "{}{}{}".format(self.table.upper(),str_filter('s',self.schema_filter),str_filter('v',self.value_filter))

    def __eq__(self, other):
        return (isinstance(other, KeywordMatch)
                and self.table == other.table
                #and set(self.keywords(schema_only=True)) == set(other.keywords(schema_only=True))
                and self.schema_filter == other.schema_filter
                and self.value_filter == other.value_filter)

    def has_default_mapping(self):
        schema_items = [attr for (attr, keywords) in self.schema_filter if attr == '*']
        value_items = [attr for (attr, keywords) in self.value_filter if attr == '*']

        return len(value_items) > 0 or len(schema_items) > 0

    def replace_default_mapping(self, attribute):
        schema_items = [attr for (attr, keywords) in self.schema_filter if attr == '*']
        value_items = [attr for (attr, keywords) in self.value_filter if attr == '*']

        if len(value_items) > 0:
            new_value_set = set({(attr, keywords) for attr, keywords in self.value_filter if attr != '*'})
            value = [keywords for (attr, keywords) in self.value_filter if attr == '*'][0]
            new_value_set.add((attribute, value))
            self.value_filter = frozenset(new_value_set)

        if len(schema_items) > 0:
            new_schema_set = set({(attr, keywords) for attr, keywords in self.schema_filter if attr != '*'})
            value = [keywords for (attr, keywords) in self.schema_filter if attr == '*'][0]
            new_schema_set.add((attribute, value))
            self.schema_filter = frozenset(new_schema_set)


    def has_same_attribute(self, other):
        all_equal_attributes = False
        attrib = set([attrib for (attrib,keywords) in self.value_filter])
        other_attrib = set([attrib for (attrib,keywords) in other.value_filter])

        if attrib != other_attrib:
            return False


        attrib = set([attrib for (attrib,keywords) in self.schema_filter])
        other_attrib = set([attrib for (attrib,keywords) in other.schema_filter])

        if attrib == other_attrib:
            all_equal_attributes=True

        return all_equal_attributes

    def __hash__(self):
        # return hash( (self.table,frozenset(self.keywords(schema_only=True)),self.value_filter) )
        return hash( (self.table,self.schema_filter,self.value_filter) )

    def to_json_serializable(self):

        def filter_object(fltr):
            filter_obj_list = []
            for attribute, keywords in fltr:
                filter_obj = {}
                filter_obj['attribute'] = attribute

                keywords_list = []

                for keyword in keywords:
                    if isinstance(keyword, frozenset):
                        for term in keyword:
                            keywords_list.append(term)
                    else:
                        keywords_list.append(keyword)
                
                filter_obj['keywords'] = keywords_list

                filter_obj_list.append(filter_obj)

            return filter_obj_list #[{'attribute':attribute,
                            #'keywords':list(keywords)} for attribute,keywords in fltr]

        return {'table':self.table,
                'schema_filter':filter_object(self.schema_filter),
                'value_filter':filter_object(self.value_filter),}

    def to_json(self):
        return json.dumps(self.to_json_serializable())

    @staticmethod
    def from_str(str_km):
        re_km = re.compile(r'([A-Z,_,1-9]+)(.*)')
        re_filters = re.compile(r'\.([vs])\(([^\)]*)\)')
        re_predicates = re.compile(r'([\w\*]*)\{([^\}]*)\}\,?')
        re_keywords = re.compile(r'(\w+)\,?')

        m_km=re_km.match(str_km)
        table = m_km.group(1).lower()
        schema_filter={}
        value_filter={}
        for filter_type,str_predicates in re_filters.findall(m_km.group(2)):

            if filter_type == 'v':
                predicates = value_filter
            else:
                predicates = schema_filter
            for attribute,str_keywords in re_predicates.findall(str_predicates):
                predicates[attribute]={key for key in re_keywords.findall(str_keywords)}
        return KeywordMatch(table,value_filter=value_filter,schema_filter=schema_filter,)

    @staticmethod
    def from_json_serializable(json_serializable):

        # print(json_serializable)

        def filter_dict(filter_obj):
            filter_dict_obj = {}
            
            for predicate in filter_obj:
                filter_dict_obj[predicate['attribute']] = []
                
                for keyword in predicate['keywords']:
                    if isinstance(keyword, frozenset):
                        for term in keyword:
                            filter_dict_obj[predicate['attribute']].append(
                                term)
                    else:
                        filter_dict_obj[predicate['attribute']].append(keyword)
            return filter_dict_obj

        return KeywordMatch(json_serializable['table'],
                            value_filter  = filter_dict(json_serializable['value_filter']),
                            schema_filter  = filter_dict(json_serializable['schema_filter']))
    @staticmethod
    def from_json(str_json):
        return KeywordMatch.from_json_serializable(json.loads(str_json))



import json
from timeit import default_timer as timer
from pprint import pprint as pp
import itertools

from nltk.data import retrieve
from nltk.corpus import wordnet as wn

from k2db.config import DefaultConfiguration
from k2db.database import DatabaseHandler, MongoHandler, MongoQueryExecutor
from k2db.utils import (
    ConfigHandler,
    KeywordQuery,
    Tokenizer,
    Similarity,
    get_logger,
    next_path,
)
from k2db.utils.utils import DictToSTupleFormatter, modify_retrieved_documents
from k2db.index import IndexHandler
from k2db.keyword_match import KeywordMatchHandler
from k2db.query_match import QueryMatchHandler
from k2db.candidate_network import CandidateNetworkHandler
from k2db.evaluation import EvaluationHandler

s = Similarity('')

logger = get_logger(__name__)
class Mapper:

    DATABASE_HANDLERS = {
        'mongo': MongoHandler,
        'postgres': DatabaseHandler,
    }

    def __init__(self, config = None):
        self.config = config
        if self.config is None:
            self.config = ConfigHandler()

        self.database_handler = self.DATABASE_HANDLERS.get(
            DefaultConfiguration.DATABASE_ENGINE
            )(DefaultConfiguration.DATABASE)
        self.index_handler = IndexHandler(self.database_handler)

        self.tokenizer = Tokenizer(tokenize_method = 'simple')
        self.index_handler.load_indexes()
        # print(f'SCHEMA GRAPH:\n{self.index_handler.schema_graph.str_edges_info()}')

        self.similarity = Similarity(self.index_handler.schema_index)
        self.keyword_match_handler = KeywordMatchHandler(self.similarity)
        self.query_match_handler = QueryMatchHandler()
        self.evaluation_handler = EvaluationHandler()
        self.candidate_network_handler = CandidateNetworkHandler(
            database_handler = self.database_handler
        )
        self.evaluation_handler.load_golden_standards()

    def load_queryset(self):
        with open(self.config.queryset_filepath,mode='r') as f:
            queryset = json.load(f)
        self.queryset = queryset

    def run_queryset(self,**kwargs):
        '''
        results_filename is declared here for sake of readability. But the
        default value is only set later to get a timestamp more accurate.
        '''
        results_filename = kwargs.get('results_filename',None)
        export_results = kwargs.get('export_results',False)
        approach = kwargs.get('approach','standard')
        preprocessed_results = kwargs.get('preprocessed_results',{})

        results =[]

        keywords_to_load = {keyword for item in self.queryset for keyword in set(self.tokenizer.keywords(item['keyword_query']))}

        self.index_handler.load_indexes(keywords = keywords_to_load)

        kwargs['queries_list'] = []

        for query_number, item in enumerate(self.queryset):
            keyword_query = item['keyword_query']
            kwargs['query_number'] = query_number

            # if query_number != 2:
            #     continue

            print(f'Running keyword search for query: {keyword_query}')
            if keyword_query in preprocessed_results:
                print('  Preprocessed results loaded')
                result = preprocessed_results[keyword_query]
            else:
                result = self.keyword_search(keyword_query,**kwargs)
            results.append(result)
        
        with open('generated_queries.json', 'w') as f:
            json.dump(kwargs['queries_list'], f, indent=2)

        data = {
            "database":self.config.connection['database'],
            "queryset":self.config.queryset_filepath,
            "results":results,
        }

        if export_results:
            if results_filename is None:
                results_filename = next_path(f'{self.config.results_directory}{self.config.queryset_name}-{approach}-%03d.json')

            with open(results_filename,mode='w') as f:
                logger.info(f'Writing results in {results_filename}')
                json.dump(data,f, indent = 4)

        return data


    def keyword_search(self,keyword_query,**kwargs):
        repeat = kwargs.get('repeat',1)
        assume_golden_qms = kwargs.get('assume_golden_qms',False)
        max_num_query_matches = kwargs.get('max_num_query_matches',5)
        input_desired_cn = kwargs.get('input_desired_cn',False)
        skip_cn_generations = kwargs.get('skip_cn_generations',False)

        # s.word_similarity('test', 'test', 'test')

        weight_scheme = kwargs.get('weight_scheme',3)
        #preventing to send multiple values for weight_scheme
        if 'weight_scheme' in kwargs:
            del kwargs['weight_scheme']

        elapsed_time = {
            'km':[],
            'skm':[],
            'vkm':[],
            'qm':[],
            'cn':[],
            'total':[],
        }

        keywords = self.tokenizer.extract_keywords(keyword_query)
        keyword_query = KeywordQuery(keywords, keyword_query)

        for _ in range(repeat):
            if not assume_golden_qms:
                start_skm_time = timer()
                sk_matches = self.keyword_match_handler.schema_keyword_match_generator(
                    keyword_query, self.index_handler.schema_index)
                logger.info('%d SKMs generated: %s',len(sk_matches),sk_matches)

                start_vkm_time = timer()
                vk_matches = self.keyword_match_handler.value_keyword_match_generator(
                    keyword_query, self.index_handler.value_index)
                # vk_matches = self.keyword_match_handler.filter_kwmatches_by_compound_keywords(vk_matches,compound_keywords)
                logger.info('%d VKMs generated: %s',len(vk_matches),vk_matches)

                start_qm_time = timer()
                print('Elapsed VKM: {}'.format(start_qm_time - start_vkm_time))
                kw_matches = sk_matches + vk_matches

                pp(kw_matches)

                # exit(1)

                query_matches = self.query_match_handler.generate_query_matches(
                    keyword_query.get_keywords(), kw_matches, **kwargs)
            else:
                start_skm_time = timer()
                start_vkm_time = timer()
                start_qm_time = timer()
                kw_matches = []
                query_matches = self.evaluation_handler.golden_standards[keyword_query]['query_matches']

            # print('Keyword Matches')
            # pp(kw_matches)

            ranked_query_matches = self.query_match_handler.rank_query_matches(query_matches,
                self.index_handler.value_index,
                self.index_handler.schema_index,
                self.similarity,
                weight_scheme,
            )

            # print('Query Matches')
            # for n, qm in enumerate(ranked_query_matches[:5]):
            #     print('QM #{}'.format(n))
            #     pp(qm)

            ranked_query_matches = ranked_query_matches[:max_num_query_matches]        

            logger.info('%d QMs generated: %s',len(ranked_query_matches),ranked_query_matches)

            if input_desired_cn:
                desired_cn = self.evaluation_handler.golden_standards[keyword_query]['candidate_networks'][0]
                kwargs['desired_cn'] = desired_cn
            else:
                kwargs['desired_cn'] = None

            if not skip_cn_generations:
                start_cn_time = timer()

                ranked_cns = self.candidate_network_handler.generate_cns(
                    self.index_handler.schema_index,
                    self.index_handler.schema_graph,
                    ranked_query_matches,
                    keyword_query.get_keywords(),
                    weight_scheme,
                    **kwargs,
                )

                for n, cn in enumerate(ranked_cns[:5]):
                    print('CN #{}'.format(n))
                    pp(cn)

                # exit(0)

                # logger.info('%d CNs generated: %s',len(ranked_cns),[(cn.score,cn) for cn in ranked_cns])
                end_cn_time = timer()

                retrieved_documents = []
                query_gen_time = None
                query_exec_time = None

                # pp(ranked_cns)
                top_cn_validation_size = kwargs.get('top_cn_validation_size', 1)

                for n, cn in enumerate(ranked_cns[:top_cn_validation_size]):
                    print(f"CN #{n}")
                    pp(cn)
                    # pp(cn.get_mongo_query_from_cn(self.index_handler.schema_graph).build())
                    if 'queries_list' in kwargs:
                        generate_query_start_time = timer()
                        base_collection, generated_query = cn.get_mongo_query_from_cn(
                            self.index_handler.schema_graph
                        ).build()
                        pp(generated_query)
                        # continue
                        generate_query_end_time = timer()
                        kwargs['queries_list'].append(
                            {
                                'query_number': kwargs['query_number'] + 1 if 'query_number' in kwargs else None,
                                'input_query': keyword_query.get_value(),
                                'base_collection': base_collection,
                                'mongo_query': generated_query,
                            }
                        )
                        query_executor = MongoQueryExecutor()
                        execute_query_start_time = timer()
                        retrieved_documents.extend(query_executor.execute(base_collection, generated_query))
                        execute_query_end_time = timer()

            else:
                ranked_cns=[]
            
            cn_count = len(ranked_cns)      

            elapsed_time['km'].append(   start_qm_time -start_skm_time)
            elapsed_time['skm'].append(  start_vkm_time-start_skm_time)
            elapsed_time['vkm'].append(  start_qm_time -start_vkm_time)
            elapsed_time['qm'].append(   start_cn_time -start_qm_time)
            elapsed_time['cn'].append(   end_cn_time   -start_cn_time)
            elapsed_time['total'].append(end_cn_time   -start_skm_time)

            print(elapsed_time['skm'])
            print(elapsed_time['vkm'])

        aggregated_elapsed_time = {phase:min(times) for phase,times in elapsed_time.items()}

        if retrieved_documents:
            query_gen_time = generate_query_end_time - generate_query_start_time
            query_exec_time = execute_query_end_time - execute_query_start_time
            modified_documents = modify_retrieved_documents(list(retrieved_documents))

        result = {
            'keyword_query':keyword_query.get_value(),
            'keywords':list(keywords),
            'query_number': kwargs['query_number'],
            # 'keyword_matches': [kw_match.to_json_serializable() 
                                # for kw_match in kw_matches],
            # 'compound_keywords':list(compound_keywords),
            'query_matches':      [query_match.to_json_serializable()
                                  for query_match in ranked_query_matches],
            'candidate_networks': [candidate_network.to_json_serializable()
                                  for candidate_network in ranked_cns],
            'generated_query': kwargs['queries_list'][-1]['mongo_query'] if 'queries_list' in kwargs else None,
            'retrieved_documents': modified_documents if retrieved_documents else retrieved_documents,
            'elapsed_time':       aggregated_elapsed_time,
            'num_keyword_matches':len(kw_matches),
            #consider len of unranked query matches
            'num_query_matches':  len(query_matches),
            'num_candidate_networks':  len(ranked_cns),
            'query_generation_time': query_gen_time,
            'query_execution_time': query_exec_time,
            'num_value_keyword_matches': len(vk_matches),
            'num_schema_keyword_matches': len(sk_matches),
            # 'qm_count': qm_count,
            # 'cn_count': cn_count,
        }

        # print('retrieved_documents')
        # pp(result['retrieved_documents'])

        return result



class Graph:
    __slots__ = ['_graph_dict','_edges_info']
    def __init__(self, graph_dict=None, edges_info=None):
        if graph_dict is None:
            graph_dict = {}
        self._graph_dict = graph_dict

        if edges_info is None:
            edges_info = {}
        self._edges_info = edges_info

    def add_vertex(self, vertex):
        self._graph_dict.setdefault(vertex, (set(),set()) )
        return vertex

    def add_outgoing_neighbor(self,vertex,neighbor):
        self._graph_dict[vertex][0].add(neighbor)

    def add_incoming_neighbor(self,vertex,neighbor):
        self._graph_dict[vertex][1].add(neighbor)

    def outgoing_neighbors(self,vertex):
        yield from self._graph_dict[vertex][0]

    def incoming_neighbors(self,vertex):
        yield from self._graph_dict[vertex][1]

    def neighbors(self,vertex):
        #This method does not use directed_neighbors to avoid generating tuples with directions
        yield from self.outgoing_neighbors(vertex)
        yield from self.incoming_neighbors(vertex)

    def directed_neighbors(self,vertex):
        #This method does not use directed_neighbors to avoid generating tuples with directions
        for outgoing_neighbor in self.outgoing_neighbors(vertex):
            yield ('>',outgoing_neighbor)
        for incoming_neighbor in self.incoming_neighbors(vertex):
            yield ('<',incoming_neighbor)

    def add_edge(self, vertex1, vertex2,edge_info = None, edge_direction='>'):
        if edge_direction=='>':
            self.add_outgoing_neighbor(vertex1,vertex2)
            self.add_incoming_neighbor(vertex2,vertex1)
            if edge_info is not None:
                self._edges_info[(vertex1, vertex2)] = edge_info

        elif edge_direction=='<':
            self.add_edge(vertex2, vertex1,edge_info = edge_info)
        else:
            raise SyntaxError('edge_direction must be > or <')

    def get_edge_info(self,vertex1,vertex2):
        return self._edges_info.get( (vertex1, vertex2), None )

    def vertices(self):
        return self._graph_dict.keys()

    def edges(self):
            for vertex in self.vertices():
                for neighbor in self.outgoing_neighbors(vertex):
                    yield (vertex,)+ (neighbor,)

    def dfs_pair_iter(self, source_iter=None, root_predecessor=False, rev=False, **kwargs):
        last_vertex_by_level=[]

        if source_iter is None:
            source_iter = self.leveled_dfs_iter(rev=rev, **kwargs)

        for direction,level,vertex in source_iter:
            if level < len(last_vertex_by_level):
                last_vertex_by_level[level] = vertex
            else:
                last_vertex_by_level.append(vertex)

            if level>0:
                prev_vertex = last_vertex_by_level[level-1]
                yield (prev_vertex,direction,vertex)
            elif root_predecessor:
                yield (None,'',vertex)


    def leveled_dfs_iter(self,start_vertex=None,visited = None, level=0, direction='',two_way_transversal=True, rev=False):
        if len(self)>0:
            if start_vertex is None:
                start_vertex = self.get_starting_vertex(rev)
            if visited is None:
                visited = set()
            visited.add(start_vertex)

            yield( (direction,level,start_vertex) )

            for neighbor in self.outgoing_neighbors(start_vertex):
                if neighbor not in visited:
                    yield from self.leveled_dfs_iter(neighbor,visited,
                                                     level=level+1,
                                                     direction='>',
                                                     two_way_transversal=two_way_transversal,
                                                     rev=rev)

            # two_way_transversal indicates whether the DFS will expand through incoming neighbors
            if two_way_transversal:
                for neighbor in self.incoming_neighbors(start_vertex):
                    if neighbor not in visited:
                        yield from self.leveled_dfs_iter(neighbor,visited,
                                                         level=level+1,
                                                         direction='<',
                                                         two_way_transversal=two_way_transversal,
                                                         rev=rev)


    def leaves(self):
        for vertice in self.vertices():
            if sum(1 for neighbor in self.neighbors(vertice)) == 1:
                yield vertice

    def num_leaves(self):
        return len(list(self.leaves()))

    def get_starting_vertex(self, rev):
        vertices = iter(self.vertices())
        if rev:
            return vertices[-1]
        return next(vertices)

    def __repr__(self):
        return repr(self._graph_dict)

    def __len__(self):
         return len(self._graph_dict)

    def str_graph_dict(self):
        return str(self._graph_dict)

    def str_edges_info(self):
        return str(self._edges_info)



import json
import logging
import re
from glob import glob
from pprint import pprint as pp


debug_mapping = { \
	'DEBUG' : logging.DEBUG,\
	'INFO': logging.INFO, \
	'ERROR': logging.ERROR, \
	'WARNING': logging.WARNING, \
	'CRITICAL': logging.CRITICAL \
}

class ConfigHandler:
    __instance = None

    config_folder_directory = None
    connection = {}
    logging_mode = None
    create_index = None
    results_directory = None
    plots_directory = None
    queryset_config_filepath = None
    dataset_config_filepath = None
    queryset_filepath = None
    queryset_name = None
    attributes_filepath = None
    schema_index_filepath = None
    value_index_filepath = None
    schema_graph_filepath = None

    def __init__(self, config_directory, reset=False, **kwargs):

        if ConfigHandler.__instance is None or reset:
            config_folder_directory = config_directory
            general_config_file = f'{config_folder_directory}config.json'

            config = self.load_config(general_config_file)
            config['logging_mode'] = debug_mapping[config['logging_mode']]

            if 'queryset_config_filepath' in kwargs:
                config['queryset_config_filepath'] = kwargs['queryset_config_filepath']

            queryset_config = self.load_config(config['queryset_config_filepath'])

            if 'dataset_config_filepath' in kwargs:
                queryset_config['dataset_config_filepath'] = kwargs['dataset_config_filepath']         

            dataset_config = self.load_config(queryset_config['dataset_config_filepath'])

            config['connection']['database'] = dataset_config['database']
            del dataset_config['database']

            ConfigHandler.__instance = config
            ConfigHandler.__instance.update(queryset_config)
            ConfigHandler.__instance.update(dataset_config)
        
        # from pprint import pprint as pp
        # pp(ConfigHandler.__instance)

        self.__dict__ = ConfigHandler.__instance

    def update_paths(self, config, prefix_directory):
        re_path = re.compile(".+(_filepath|_directory)")
        for key in config:
            #keys that contains path info and are not None
            if config[key] and re_path.match(key):
                config[key] = config[key]
                # config[key]=prefix_directory+config[key]

    def load_config(self, filepath):
        with open(filepath, 'r') as f:
            config = json.load(f)
        return config

    def set_logging_level(self, ĺevel):
        self.logging_mode = debug_mapping[ĺevel]
        ConfigHandler.__instance['logging_mode'] = self.logging_mode

    def get_dataset_configs(self):
        subfolder = './datasets_config/'
        results = []
        for filepath in glob(f'{self.config_folder_directory}{subfolder}*.json'):
            with open(filepath,'r') as f:
                results.append( (json.load(f)['database'], filepath) )
        return results

    def get_queryset_configs(self,dataset_config_filepath=None):
        subfolder = './querysets_config/'
        results = []
        for filepath in glob(f'{self.config_folder_directory}{subfolder}*.json'):
            with open(filepath,'r') as f:
                data = json.load(f)
                if dataset_config_filepath in (None,data['dataset_config_filepath']):
                    results.append( (data['queryset_name'], filepath) )
        return results



from pprint import pprint as pp

from graphviz import Digraph
from IPython.display import display

from sereia.candidate_network import CandidateNetwork
from sereia.query_match import QueryMatch
from sereia.keyword_match import KeywordMatch
from sereia.utils.printmd import printmd
from sereia.utils.shift_tab import shift_tab
from sereia.utils.ordinal import ordinal


class SereiaResult():

    def __init__(self, database_name, index_handler, database_handler, data):
        self.data = data
        self.database_name = database_name
        self.database_handler = database_handler
        self.index_handler = index_handler

    def cjns(self, **kwargs):
        show_str = kwargs.get('show_str', kwargs.get('text', True))
        show_graph = kwargs.get('show_graph', kwargs.get('graph', True))
        show_sql = kwargs.get('show_sql', kwargs.get('sql', True))
        show_df = kwargs.get('show_df', kwargs.get(
            'df', kwargs.get('jnts', True)))
        top_k = kwargs.get('top_k', 0)
        head = kwargs.get('head', 5)

        for i, json_cn in enumerate(self.data['candidate_networks']):
            cjn = CandidateNetwork.from_json_serializable(json_cn)
            printmd('---')
            printmd(f'**{ordinal(i+1)} CJN**:')

            if show_str:
                printmd('---')
                print(f'Text:\n{shift_tab(cjn)}')
            if show_graph:
                g = Digraph(
                    graph_attr={'nodesep': '0.2', 'ranksep': '0.25'},
                    node_attr={'fontsize': "9.0", },
                    edge_attr={'arrowsize': '0.9', },
                )
                for label, id in cjn.vertices():
                    g.node(id, label=str(label))
                for (label_a, id_a), (label_b, id_b) in cjn.edges():
                    g.edge(id_a, id_b)
                printmd('---')
                print('Graph:')
                display(g)
            if show_df or show_sql:
                mongo_query = cjn.get_mongo_query_from_cn(
                    self.database_name,
                    self.index_handler.schema_graph,
                ).build()[1]
                
                if show_sql:
                    printmd('---')
                    print('Mongo Query:')
                    pp(mongo_query)
                # if show_df:
                #     printmd('---')
                #     print(f'JNTs:')
                #     df = self.database_handler.get_dataframe(sql)
                #     if head > 0:
                #         df = df.head(head)
                #     if df.empty:
                #         print('\tThis CJN returns no tuples.')
                #     else:
                #         df_hide_tsvector = df.loc[:, ~
                #                                   df.columns.str.endswith('tsvector')]
                #         # dt = data_table.DataTable(df_hide_tsvector, include_index=False, num_rows_per_page=5)
                #         # display(dt)
                #         display(df_hide_tsvector)
            print()

            if i+1 >= top_k and top_k > 0:
                break

    def qms(self, **kwargs):
        top_k = kwargs.get('top_k', 0)

        for i, json_qm in enumerate(self.data['query_matches']):
            qm = QueryMatch.from_json_serializable(json_qm)

            printmd('---')
            printmd(f'**{ordinal(i+1)} QM**:')
            printmd('---')
            print(qm, end='\n')

            if i+1 >= top_k and top_k > 0:
                break

    def kms(self, **kwargs):
        show_skms = kwargs.get('schema', True)
        show_vkms = kwargs.get('value', True)
        top_k = kwargs.get('top_k', 0)

        self.skms(top_k=top_k)
        self.vkms(top_k=top_k)

    def skms(self, **kwargs):
        top_k = kwargs.get('top_k', 0)
        printmd('---')
        printmd(f'**SKMs**:')
        printmd('---')

        if len(self.data['schema_keyword_matches']) == 0:
            print('There is no SKM for this keyword query.')
            return
        for i, json_km in enumerate(self.data['schema_keyword_matches']):
            km = KeywordMatch.from_json_serializable(json_km)
            #print(f'{i+1} KM:')
            print(km)
            if i+1 >= top_k and top_k > 0:
                break

    def vkms(self, **kwargs):
        top_k = kwargs.get('top_k', 0)
        printmd('---')
        printmd(f'**VKMs**:')
        printmd('---')
        if len(self.data['value_keyword_matches']) == 0:
            print('There is no VKM for this keyword query.')
            return
        for i, json_km in enumerate(self.data['value_keyword_matches']):
            km = KeywordMatch.from_json_serializable(json_km)
            print(km)
            if i+1 >= top_k and top_k > 0:
                break

    def __repr__(self):
        return 'SereiaResult'



import gc
from glob import glob
import shelve
from contextlib import ExitStack
import json
from pprint import pprint as pp
from math import log
from time import time
import os

from sereia.utils import (
    calculate_tf,
    calculate_inverse_frequency,
    calculate_iaf,
    ConfigHandler,
    load_schema_graph_from_json,
    memory_size,
    load_database_structure_from_file,
)

from sereia.index.value_index import ValueIndex
from sereia.index.schema_index import SchemaIndex
from sereia.index.schema_graph import SchemaGraph
from sereia.index.babel_hash import BabelHash


class IndexHandler:

    def __init__(self, database_name, database_handler, config_handler, **kwargs):
        self.config = config_handler
        self.value_index = ValueIndex()
        self.schema_index = SchemaIndex()
        self.schema_graph = SchemaGraph()
        self.database_name = database_name
        self.database_handler = database_handler
        self.partial_index_count = 0

    def create_indexes(self):
        if not self.config.create_index:
            print(f'Index Creation is disabled.')
            return

        self.create_schema_graph()
        self.create_partial_schema_index()
        self.create_partial_value_indexes()
        self.merge_partial_value_indexes_and_process_max_frequency()
        self.process_norms()

    def create_schema_graph(self):
        fk_constraints = load_schema_graph_from_json(self.database_name)
        self.database_handler.get_collections_names()

        for constraint,values in fk_constraints.items():
            self.schema_graph.add_fk_constraint(constraint,*values)

        if not len(self.schema_graph):
            self.schema_graph.add_root(
                self.database_handler.get_collections_names()[0],
            )

        self.schema_graph.persist_to_file(
            self.config.dataset_directory,
            self.config.schema_graph_filepath,
        )

    def create_partial_schema_index(self):
        tables_attributes = None

        tables_attributes = self.database_handler.get_collections_and_attributes()
        self.database_schema_elements = tables_attributes

        metrics = {'max_frequency':0, 'norm':[0,0,0,0]}
        self.schema_index.create_entries(tables_attributes,metrics)

    def create_partial_value_indexes(self,**kwargs):
        unit = kwargs.get('unit', 2**30)
        max_memory_allowed = kwargs.get('max_memory', 4)
        max_memory_allowed *= unit
        gb = 1 * unit

        indexable_attributes = self.database_schema_elements
        attributes_filepath = self.config.attributes_filepath
        if os.path.exists(attributes_filepath):
            print('Indexable attributes file exists')
            with open(attributes_filepath,'r') as indexable_attributes_file:     
                indexable_attributes = [
                    (item['table'],attribute)
                    for item in json.load(indexable_attributes_file)
                    for attribute in item['attributes']
                ]
        else:
            database_structure = load_database_structure_from_file(self.database_name)
            print('Indexable attributes file does not exist')
            indexable_attributes = [
                (collection, attribute)
                for collection in database_structure
                for attribute in database_structure[collection]
                if database_structure[collection][attribute] not in [type(dict()), type(list())]
            ]

        def part_index():
            self.partial_index_count += 1
            partial_index_filename = f'{self.config.value_index_filepath}.part{self.partial_index_count:02d}'
            self.value_index.persist_to_file(self.config.dataset_directory, partial_index_filename)
            self.value_index = ValueIndex()
            gc.collect()

        for table, ctid, attribute, word in self.database_handler.iterate_over_keywords(
                self.schema_index,
                indexable_attributes,
            ):
            self.value_index.add_mapping(word,table,attribute,ctid)

            if memory_size() >= max_memory_allowed:
                print(f'Memory usage exceeded the maximum memory allowed (above {max_memory_allowed/gb:.2f}GB).')
                part_index()

        part_index()

    def merge_partial_value_indexes_and_process_max_frequency(self):
        num_total_attributes=self.schema_index.get_num_total_attributes()
        babel = BabelHash.babel

        value_index_path = self.config.dataset_directory + self.config.value_index_filepath
        filenames=glob(f'{value_index_path}.part*')
        filenames = [file.replace('.db', '') for file in filenames]

        with ExitStack() as stack:
            partial_value_indexes = [stack.enter_context(shelve.open(fname,flag='r')) for fname in filenames]
            final_value_index = stack.enter_context(shelve.open(f'{value_index_path}'))

            for partial_value_index in partial_value_indexes:
                babel.update(partial_value_index['__babel__'])
            final_value_index['__babel__'] = babel

            for i in range(len(partial_value_indexes)):
                for word in partial_value_indexes[i]:
                    if word in final_value_index or word == '__babel__':
                        continue

                    value_list = [ partial_value_indexes[i][word] ]

                    for j in range(i+1,len(partial_value_indexes)):
                        if word in partial_value_indexes[j]:
                            value_list.append(partial_value_indexes[j][word])

                    merged_babel_hash = BabelHash()
                    if len(value_list) > 1:
                        for ( _ ,part_babel_hash) in value_list:
                            for table in part_babel_hash:
                                merged_babel_hash.setdefault(table,BabelHash())
                                for attribute in part_babel_hash[table]:
                                    merged_babel_hash[table].setdefault( attribute , [] )
                                    merged_babel_hash[table][attribute]+= part_babel_hash[table][attribute]
                    else:
                        _ ,merged_babel_hash = value_list[0]

                    for table in merged_babel_hash:
                        for attribute in merged_babel_hash[table]:
                            frequency = len(merged_babel_hash[table][attribute])
                            max_frequency = self.schema_index[table][attribute]['max_frequency']
                            if frequency > max_frequency:
                               self.schema_index[table][attribute]['max_frequency'] = frequency


                    num_attributes_with_this_word = sum([len(merged_babel_hash[table]) for table in merged_babel_hash])

                    inverse_frequency = calculate_inverse_frequency(num_total_attributes,num_attributes_with_this_word)
                    merged_value = (inverse_frequency,merged_babel_hash)
                    final_value_index[word]=merged_value

    def process_norms(self):
        value_index_path = self.config.dataset_directory + self.config.value_index_filepath
        with shelve.open(f'{value_index_path}') as full_index:
            for word in full_index:
                if word == '__babel__':
                    continue

                inverse_frequency, babel_hash = full_index[word]
                for table in babel_hash:
                    for attribute in babel_hash[table]:
                        frequency = len(babel_hash[table][attribute])
                        max_frequency = self.schema_index[table][attribute]['max_frequency']
                        
                        for weight_scheme in range(4):
                            tf  = calculate_tf(weight_scheme,frequency,max_frequency)
                            iaf = calculate_iaf(weight_scheme,inverse_frequency)
                            self.schema_index[table][attribute]['norm'][weight_scheme] += (tf*iaf)**2

        for table in self.schema_index:
            for attribute in self.schema_index[table]:
                for weight_scheme in range(4):
                    self.schema_index[table][attribute]['norm'][weight_scheme] **= 0.5

        self.schema_index.persist_to_file(
            self.config.dataset_directory,
            self.config.schema_index_filepath,
        )

    def load_indexes(self,**kwargs):
        schema_graph_path = self.config.dataset_directory + self.config.schema_graph_filepath
        value_index_path = self.config.dataset_directory + self.config.value_index_filepath
        schema_index_path = self.config.dataset_directory + self.config.schema_index_filepath

        self.schema_graph.load_from_file(schema_graph_path)
        self.value_index.load_from_file(value_index_path, **kwargs)
        self.schema_index.load_from_file(schema_index_path)



from random import sample
import json

from sereia.keyword_match import KeywordMatch


class QueryMatch:
    '''
    Query match is a set of tuple-sets that, if properly joined,
    can produce networks of tuples that fulfill the query. They
    can be thought as the leaves of a Candidate Network.
    '''

    def __init__(self, keyword_matches):
        self.keyword_matches = keyword_matches
        # self.value_score = 1.0
        # self.schema_score = 1.0
        # self.total_score = 1.0
        # self.tables_on_match = set()

    def get_km_from_keyword(self,keyword):
        for keyword_match in self.keyword_matches:
            if keyword in keyword_match.keywords():
                return keyword_match
        return None

    def get_random_keyword_match(self):
        return sample(self.keyword_matches,k=1)[0]

    def calculate_total_score(self, value_index, schema_index, similarity, weight_scheme):
        self.total_score = 1

        has_value_terms = self.calculate_value_score(value_index, schema_index,weight_scheme)
        has_schema_terms = self.calculate_schema_score(similarity)

        if has_value_terms:
            self.total_score *= self.value_score

        if has_schema_terms:
            self.total_score *= self.schema_score

        self.total_score /= len(self)

    def calculate_schema_score(self,similarity):
        self.schema_score = 1
        has_schema_terms = False
        for keyword_match in self.keyword_matches:
            for table, attribute, schema_words in keyword_match.schema_mappings():
                has_schema_terms = True
                sim_values = [
                    similarity.word_similarity(term,table,attribute) 
                    for term in schema_words
                ]
                self.schema_score *= sum(sim_values)/len(sim_values)
        return has_schema_terms

    def calculate_value_score(self, value_index, schema_index, weight_scheme):
        has_value_terms = False
        self.value_score = 1

        for keyword_match in self.keyword_matches:
            for table, attribute, keywords in keyword_match.value_mappings():
                has_value_terms = True
                metrics = schema_index[table][attribute]
                tf_iaf_sum = 0
                for keyword in keywords:
                    if isinstance(keyword, frozenset):
                        for term in keyword:
                            tf = value_index.get_tf(weight_scheme,term,table,attribute,metrics['max_frequency'])
                            iaf = value_index.get_iaf(weight_scheme, term)
                            tf_iaf_sum += (tf*iaf)
                    else:
                        tf = value_index.get_tf(weight_scheme,keyword,table,attribute,metrics['max_frequency'])
                        iaf = value_index.get_iaf(weight_scheme, keyword)
                        tf_iaf_sum+= (tf*iaf)

                cos = tf_iaf_sum / metrics['norm'][weight_scheme]
                self.value_score *= cos
        return has_value_terms

    def __eq__(self, other):
        return (isinstance(other, QueryMatch)
                and self.keyword_matches == other.keyword_matches)

    def __iter__(self):
        return iter(self.keyword_matches)

    def __len__(self):
        return len(self.keyword_matches)

    def __repr__(self):
        return repr(self.keyword_matches)

    def __str__(self):
        return repr(self.keyword_matches)

    def to_json_serializable(self):
        return [keyword_match.to_json_serializable() for keyword_match in self.keyword_matches]

    def to_json(self):
        return json.dumps(self.to_json_serializable())

    @staticmethod
    def from_json_serializable(json_serializable_qm):
        return QueryMatch({KeywordMatch.from_json_serializable(json_serializable_km)
                           for json_serializable_km in json_serializable_qm})

    @staticmethod
    def from_json(str_json):
        return QueryMatch.from_json_serializable(json.loads(str_json))
